/*
Copyright (c) 1992 Sandia National Laboratories

  Redistribution and use in source and binary forms are permitted
  provided that source distributions retain this entire copyright
  notice.  Neither the name of Sandia National Laboratories nor
  the name of the author may be used to endorse or promote products
  derived from this software without specific prior written permission.

Warranty:
  This software is provided "as is" and without any express or
  implied warranties, including, without limitation, the implied
  warranties of merchantability and fitness for a particular
  purpose.

Package title:
  Out-of-core LU factorization for nCUBE (Vertex)

General Description:
  This code computes the LU factorization of a matrix on an nCUBE
  parallel computer running the Vertex operating system
  The matrices L is not available explicitly, but rather as a 
  product of Gauss matrices and permutation matrices.

Author: 
  David E. Womble
  Sandia National Laboratories, Dept. 1422
  P.O. Box 5800
  Albuquerque, NM 87185-5800

  dewombl@cs.sandia.gov
  (505) 845-7471

Author's conditions of release:
  1. The author received acknowledgement in any publication or in 
     any distributions of code in which this code is used
  2. The author is notified of any problems encountered with the
     code

Notes:
  Notes on the design, behavior and use of the program are in the
  README file.
*/

#include <math.h>

/* This function returns the inverse grey code of a number */

inv_grey_c(p)
int p;
{
  int k;
  for(k = p>>1; k != 0; k >>= 1)
    p ^= k;
  return(p);
}

/* This function does a fanout to the subcube specified by a mask.
   Logical processor 0 starts the fanout */

fanout(idnode, mask, buf, mlen)
int idnode;
int *buf;
int mask, mlen;
{
  int mnode, mflag;
  if ((idnode & mask) != 0){
    mnode = -1;
    mask = -1;
    nread(buf, mlen, &mnode, &mask, &mflag);
  }
  broadcast(idnode, mask, buf, mlen);
}

/* This function does a fanout to the subcube specified by a mask.
   The processor specified in the calling argument list starts the fanout */

fanout1(idnode, mask, buf, mlen, sproc)
int idnode;
int *buf;
int mask, mlen, sproc;
{
  int mnode, lnode, mflag;
  lnode = idnode^sproc;
  if ((lnode & mask) != 0){
    mnode = -1;
    mask = -1;
    nread(buf, mlen, &mnode, &mask, &mflag);
  }
  broadcast(idnode, mask, buf, mlen);
}

/* This function does a fanout to the subcube specified by a mask.
   The processor specified in the calling argument list starts the fanout.
   Every processor must know the mask and the starting processor. */

fanout2(me, mask, buf, mlen, root, type)
int *buf;
int me, mask, mlen, root, type;
{
  int dest, check, cflag;
  for (check=1; check<=mask; check<<=1)
    if (mask & check)
      if ((me & check) != (root & check)) break;
  if (me != root) {
    dest = me ^ check;
    nread(buf, mlen, &dest, &type, &cflag);
  }
  for (check>>=1; check!=0; check>>=1)
    if (mask & check) {
      dest = me ^ check;
      nwrite(buf, mlen, dest, type, &cflag);
    }
}

/* This function sends a message along each channel specified by a mask. */

broadcast(idnode, mask, buf, mlen)
int idnode;
int *buf;
int mask, mlen;
{
  int dmask, mflag;
  for(dmask = 1; dmask <= mask; dmask = dmask<<1);
  dmask = dmask>>1;
  while((dmask & mask) != 0) {
    mask = mask^dmask;
    nwrite(buf, mlen, dmask^idnode, mask, &mflag);
    while(((dmask & mask) == 0) && dmask != 0) dmask = dmask>>1;
  }
}

/* This function does a binary exchange over a subcube specified by a mask.
   The output is the maximum of all numbers exchanged. */

binex(idnode, mask, buf, mlen)
int idnode;
int *buf;
int mask, mlen;
{
  int dmask, mflag, mnode, mtype, temp;
  for(dmask = 1; dmask <= mask; dmask = dmask<<1);
  dmask = dmask>>1;
  while((dmask & mask) != 0) {
    mnode = dmask^idnode;
    mtype = 1<<10;
    nwrite(buf, mlen, mnode, mtype, &mflag);
    nread(&temp, mlen, &mnode, &mtype, &mflag);
    *buf = ( *buf > temp ? *buf : temp);
    mask = mask^dmask;
    while(((dmask & mask) == 0) && dmask != 0) dmask = dmask>>1;
  }
}

typedef struct {
          double entry;
          int row;
        } pivot_type;

binex_pivot(idnode, mask, buf, mlen)
int idnode;
pivot_type *buf;
int mask, mlen;
{
  int dmask, mflag, mnode, mtype;
  pivot_type temp;

  for(dmask = 1; dmask <= mask; dmask = dmask<<1);
  dmask = dmask>>1;
  while((dmask & mask) != 0) {
    mnode = dmask^idnode;
    mtype = 1<<10;
    nwrite(buf, mlen, mnode, mtype, &mflag);
    nread(&temp, mlen, &mnode, &mtype, &mflag);
    if ( fabs((*buf).entry) < fabs(temp.entry)) 
      *buf = temp;
    else if ((fabs((*buf).entry) == fabs(temp.entry)) && (idnode < mnode))
      *buf = temp;
    mask = mask^dmask;
    while(((dmask & mask) == 0) && dmask != 0) dmask = dmask>>1;
  }
}
