/*
Copyright (c) 1992 Sandia National Laboratories

  Redistribution and use in source and binary forms are permitted
  provided that source distributions retain this entire copyright
  notice.  Neither the name of Sandia National Laboratories nor
  the name of the author may be used to endorse or promote products
  derived from this software without specific prior written permission.

Warranty:
  This software is provided "as is" and without any express or
  implied warranties, including, without limitation, the implied
  warranties of merchantability and fitness for a particular
  purpose.

Package title:
  Out-of-core LU factorization for nCUBE (Vertex)

General Description:
  This code computes the LU factorization of a matrix on an nCUBE
  parallel computer running the Vertex operating system
  The matrices L is not available explicitly, but rather as a 
  product of Gauss matrices and permutation matrices.

Author: 
  David E. Womble
  Sandia National Laboratories, Dept. 1422
  P.O. Box 5800
  Albuquerque, NM 87185-5800

  dewombl@cs.sandia.gov
  (505) 845-7471

Author's conditions of release:
  1. The author received acknowledgement in any publication or in 
     any distributions of code in which this code is used
  2. The author is notified of any problems encountered with the
     code

Notes:
  Notes on the design, behavior and use of the program are in the
  README file.
*/

#include <math.h>
#include <stdio.h>
#include <fcntl.h>
#include <drivers.h>

int nfd;

int create_file()
{
#include "extern_global_vars.h"

  char name[25];
  int fd, flag;
  mode_t creation_flag;

  make_file_name(me,name);

  creation_flag = 511;
  fd = open(name, O_RDWR | O_CREAT | O_TRUNC, creation_flag);
  if (fd == -1) printf("Node %d error -- can't create file %s\n",me,name);
  nfd = ioctl(fd,VERTEX_FD);

  return(fd);
}

close_file(fd)
int fd;
{
  int flag;

  flag = close(fd);
}

int get_seg(fd, seg, seg_num, bytes, rd_bytes)
int fd;
double *seg;
int seg_num, *bytes, rd_bytes;
{
#include "extern_global_vars.h"

  int i, flag;
  long offset;
  char *t;
  extern n_errno;

  offset = 0;
  for (i=0; i<seg_num; offset += bytes[i++]);

  flag = lseek(fd, offset, 0);
  if (flag < 0) {
    printf("Node %d seek error for seg %d. Error code = %d, number = %d.\n",
          me, seg_num, flag, n_errno);
  }

  flag = read(fd, seg, rd_bytes);
  if (flag != rd_bytes) {
    printf("Node %d read error for seg %d. Error code = %d, number = %d.\n",
          me, seg_num, flag, n_errno);
  }
  return (flag);
}

get_seg_st(fd, seg, seg_num, bytes, rd_bytes)
int fd;
double *seg;
int seg_num, *bytes, rd_bytes;
{
#include "extern_global_vars.h"

  int i, flag;
  long offset;
  char *t;
  extern n_errno;

  offset = 0;
  for (i=0; i<seg_num; offset += bytes[i++]);

  flag = lseek(fd, offset, 0);
  if (flag < 0) {
    printf("Node %d seek error for seg %d. Error code = %d, number = %d.\n",
          me, seg_num, flag, n_errno);
  }

  flag = n_frdst(nfd, rd_bytes);
  if (flag < 0) {
    printf("Node %d read error for seg %d. Error code = %d, number = %d.\n",
          me, seg_num, flag, n_errno);
  }
}

int get_seg_end(fd, seg, seg_num, bytes, rd_bytes)
int fd;
double *seg;
int seg_num, *bytes, rd_bytes;
{
#include "extern_global_vars.h"

  int i, flag;
  long offset;
  char *t;
  extern n_errno;

  flag = n_frdend(nfd, seg);
  if (flag != rd_bytes) {
    printf("Node %d read error for seg %d. Error code = %d, number = %d.\n",
          me, seg_num, flag, n_errno);
  }
  return(flag);
}

int put_seg(fd, seg, seg_num, bytes, wr_bytes)
int fd;
double *seg;
int seg_num, *bytes, wr_bytes;
{
#include "extern_global_vars.h"

  int i, flag;
  long offset;
  char *t;
  extern n_errno;

  offset = 0;
  for (i=0; i<seg_num; offset += bytes[i++]);

  flag = lseek(fd, offset, 0);
  if (flag < 0) {
    printf("Node %d seek error for seg %d. Error code = %d, number = %d.\n",
          me, seg_num, flag, n_errno);
  }

  flag = write(fd, seg, wr_bytes);
  if (flag != wr_bytes) {
    printf("Node %d write error for seg %d. Error code = %d, number = %d.\n",
          me, seg_num, flag, n_errno);
  }
  return(flag);
}

int hex_conv(c)
int c;
{
  c = c & 15;
  if ((c >= 0) && (c <= 9))
    return c + '0';
  else
    return c - 10 + 'a';
}

make_file_name(me, name)
int me;
char *name;
{
  char *p;
  int ndims_disk,disk_offset;

  ndims_disk = 4;
  ndims_disk = (ndims_disk < 4) ? ndims_disk : 4;
  disk_offset = 0;
  disk_offset = (((1<<ndims_disk) + disk_offset) <= 16) ? disk_offset : 0;

  p = name;
  *p++ = '/';
  *p++ = '/';
  *p++ = 'd';
  *p++ = 'f';
  *p++ = hex_conv(disk_offset + (me&((1<<ndims_disk)-1)));
  *p++ = '/';
  *p++ = 'd';
  *p++ = 'e';
  *p++ = 'w';
  *p++ = 'o';
  *p++ = 'm';
  *p++ = 'b';
  *p++ = 'l';
  *p++ = '/';
  *p++ = 'm';
  *p++ = 'a';
  *p++ = 't';
  *p++ = hex_conv(me>>(ndims_disk+4));
  *p++ = hex_conv(me>>ndims_disk);
  *p = '\0';
}
