/*
Copyright (c) 1992 Sandia National Laboratories

  Redistribution and use in source and binary forms are permitted
  provided that source distributions retain this entire copyright
  notice.  Neither the name of Sandia National Laboratories nor
  the name of the author may be used to endorse or promote products
  derived from this software without specific prior written permission.

Warranty:
  This software is provided "as is" and without any express or
  implied warranties, including, without limitation, the implied
  warranties of merchantability and fitness for a particular
  purpose.

Package title:
  Out-of-core LU factorization for nCUBE (Vertex)

General Description:
  This code computes the LU factorization of a matrix on an nCUBE
  parallel computer running the Vertex operating system
  The matrices L is not available explicitly, but rather as a 
  product of Gauss matrices and permutation matrices.

Author: 
  David E. Womble
  Sandia National Laboratories, Dept. 1422
  P.O. Box 5800
  Albuquerque, NM 87185-5800

  dewombl@cs.sandia.gov
  (505) 845-7471

Author's conditions of release:
  1. The author received acknowledgement in any publication or in 
     any distributions of code in which this code is used
  2. The author is notified of any problems encountered with the
     code

Notes:
  Notes on the design, behavior and use of the program are in the
  README file.
*/

#include <math.h>
#include <stdio.h>
#include <fcntl.h>


factor_seg(seg, seg_num)
double *seg;
int seg_num;
{
#include "extern_global_vars.h"
#include "macros.h"

  int i, j, l, k;               /* loop counters */

  int ncols_this;               /* number of cols in this segment */
  int my_cols_this;             /* number of cols I own in this seg */

  struct pivot_type {           /* a stucture for storing pivot info */
    double entry;               /*   pivot entry */
    int row;                    /*   pivit row number */
  } pivot, tpivot;              /* pivot info and some temporary pivot infor */

  int col_offset;               /* col number for first col in seg */
  int start_row;                /* row num to start column operations */
  int end_row;                  /* last row of a column operation */
  int num_rows;                 /* number of rows in a column operation */
  int num_cols;                 /* number of cols in a row operation */

  int pivot_row_in_proc;        /* local pivot row for proc containing pivot */
  int diag_row_in_proc;         /* local diag row for proc containing diag */
  int diag_col_in_proc;         /* local col number for proc containing diag */

  double *ptr1,*ptr2;           /* running ptrs into mat and update cols */
  double *ptr3, *ptr4;          /* running ptrs to diagonal and pivot rows */

  int col2_avail;               /* flag to indicate that col2 is avail. */

  double pivot_row_elem;        /* element of pivot row to use in elim. */

  int bytes;                    /* number of bytes in messages */
  int root;                     /* root processor for fanout */
  int type;                     /* mesage type for messages */
  int dest;                     /* destination processor for a single mess. */
  int flag;                     /* flag for messages */

  int one = 1;                  /* constant for the daxpy routines */
  int zero = 0;                 /* constant for the daxpy routines */

  int stride;                   /* stride for vector copy routines */
  double scaler;                /* scaler for daxpy routines */

  int error_num;                /* errors from ncube system routines */

  if (seg_num == (nsegs_row-1)) {
    ncols_this = ncols_last;
    my_cols_this = my_cols_last;
  } else {
    ncols_this = ncols_seg;
    my_cols_this = my_cols_seg;
  }

  col2_avail = -1;

  for (j=0; j<ncols_this; j++) {

    col_offset = seg_num*ncols_seg;
    start_row = (col_offset + j)/nprocs_col;
    if (my_first_row < j%nprocs_col) ++start_row;

    ptr4 = row2;
    if ((j!=0) && (me == col_owner(j-1))) ++ptr4;

    if (me == col_owner(j)) {

      if (j != 0){
        /* I have pivot column so update it  */
        ptr1 = seg + (j>>ndims_row) * (my_rows + 1) + start_row;
        ptr2 = col1 + start_row;
        num_rows = my_rows - start_row;
        scaler = -*ptr4;
        DAXPY(&num_rows, &scaler, ptr2, &one, ptr1, &one);
        ++ptr4;
      }

      /* first locate pivot entry with a search and binary exchange */

      pivot.entry = 0.0;
      ptr1 = seg + (j>>ndims_row) * (my_rows + 1) + start_row;
      num_rows = my_rows - start_row;
      if (num_rows == 1) {
        tpivot.entry = fabs((double) *ptr1);
        tpivot.row = 0;
      } else if (num_rows == 0) {
        tpivot.entry = (double) 0.0;
        tpivot.row = 0;
      } else {
        MAXMGV(ptr1, &one, &tpivot.entry, &tpivot.row, &num_rows);
      }
      if ((my_rows > 0) && (tpivot.entry >= fabs(pivot.entry))) {
        pivot.entry = *(ptr1 + tpivot.row);
        pivot.row = (tpivot.row + start_row)*nprocs_col + my_first_row;
      }
      binex_pivot(me, mask_col, &pivot, sizeof(struct pivot_type));
      *(ptr1 + num_rows) = (double) pivot.row + 0.1;
      if (pivot.entry == (double) 0.0)
        printf("Node %d error -- zero pivot found\n",me);

      /* divide everything including the diagonal by the pivot entry */

      ptr1 = seg + (j>>ndims_row) * (my_rows + 1) + start_row;
      ptr2 = ptr1;
      num_rows = my_rows - start_row;
      scaler = ((double) 1.0)/pivot.entry;
      DSCAL(&num_rows, &scaler, ptr1, &one);

      /* restore the pivot entry (which will become the diagonal) */

      if (my_first_row == pivot.row%nprocs_col) {
        pivot_row_in_proc = pivot.row/nprocs_col;
        ptr1 = seg + (j>>ndims_row)*(my_rows + 1) + pivot_row_in_proc;
        *ptr1 = pivot.entry;
      }

      /* now broadcast pivot column to other processors */

      ptr1 = seg + (j>>ndims_row)*(my_rows + 1) + start_row;
      num_rows = my_rows - start_row + 1;
      bytes = num_rows*sizeof(double);
      root = col_owner(j);
      type = 100 + seg_num;
      fanout2(me, mask_row, ptr1, bytes, root, type);
      ptr2 = col2 + start_row;
      DCOPY(&num_rows, ptr1, &one, ptr2, &one);

      col2_avail = 1;
    }

    if (j != 0) {
      diag_col_in_proc = j>>ndims_row;
      if (my_first_col <= j%nprocs_row) ++diag_col_in_proc;
      for (l=diag_col_in_proc; l<my_cols_this; l++) {
        /* update col l with col j-1 */
        ptr1 = seg + l*(my_rows + 1) + start_row;
        ptr2 = col1 + start_row;
        num_rows = my_rows - start_row;
        scaler = -*ptr4;
        DAXPY(&num_rows, &scaler, ptr2, &one, ptr1, &one);
        ++ptr4;
      }
    }

    /* If this processor is not responsible for column j, then it
       has to receive the column to use as an update */

    if (me != col_owner(j)){
      bytes = (my_rows - start_row + 1)*sizeof(double);
      root = col_owner(j);
      type = 100 + seg_num;
      ptr2 = col2 + start_row;
      fanout2(me, mask_row, ptr2, bytes, root, type);

      col2_avail = 1;
    }

      /* make sure information has been received and set the right
         pointer to it */

    if (col2_avail > 0){
      ptr2 = col1;
      col1 = col2;
      col2 = ptr2;
      col2_avail = -1;
    } else {
      printf("Node %d error -- update column not available\n",me);
    }

    /* exchange rows using information in last row of pivot column */

    pivot.row = *(col1 + my_rows);
    pivot_row_in_proc = pivot.row/nprocs_col;
    diag_col_in_proc =  j>>ndims_row;
    if (my_first_col < j%nprocs_row) ++diag_col_in_proc;
    num_cols = my_cols_this - diag_col_in_proc;
    if (pivot.row != (col_offset + j)){
      if (my_first_row == (j%nprocs_col)){
        ptr3 = seg + diag_col_in_proc*(my_rows+1) + (col_offset+j)/nprocs_col;
        stride = my_rows + 1;
        DCOPY(&num_cols, ptr3, &stride, row1, &one);
        *(row1+num_cols) = *(col1 + (j+col_offset)/nprocs_col);
        if (my_first_row != pivot.row%nprocs_col) {
          dest = row_owner(pivot.row);
          bytes = (num_cols + 1)*sizeof(double);
          type = 300;
          nwrite(row1, bytes, dest, type, &flag);
        }
      }
    }
    if (my_first_row == pivot.row%nprocs_col){
      ptr3 = seg + diag_col_in_proc*(my_rows + 1) + pivot_row_in_proc;
      stride = my_rows + 1;
      DCOPY(&num_cols, ptr3, &stride, row2, &one);
      *(row2 + num_cols) = *(col1 + pivot_row_in_proc);
    }
    /* fanout pivot row */
    root = row_owner(pivot.row);
    bytes = (my_cols_this - diag_col_in_proc + 1)*sizeof(double);
    type = 200;
    fanout2(me, mask_col, row2, bytes, root, type);
    if (pivot.row != (col_offset + j)){
      if (my_first_row == (pivot.row%nprocs_col)){
        if (my_first_row != (j%nprocs_col)) {
          dest = row_owner(j);
          bytes = (my_cols_this - diag_col_in_proc + 1)*sizeof(double);
          type = 300;
          nread(row1, bytes, &dest, &type, &flag);
        }
        ptr3 = seg + diag_col_in_proc*(my_rows + 1) + pivot_row_in_proc;
        stride = my_rows + 1;
        DCOPY(&num_cols, row1, &one, ptr3, &stride);
        *(col1+pivot_row_in_proc) = *(row1 + num_cols);
      }
      if (my_first_row == (j%nprocs_col)) {
        ptr3 = seg + diag_col_in_proc*(my_rows+1) + (col_offset+j)/nprocs_col;
        stride = my_rows + 1;
        DCOPY(&num_cols, row2, &one, ptr3, &stride);
        *(col1+(j+col_offset)/nprocs_col) = *(row2 + num_cols);
      }
    }
  }
}
