/*
Copyright (c) 1992 Sandia National Laboratories

  Redistribution and use in source and binary forms are permitted
  provided that source distributions retain this entire copyright
  notice.  Neither the name of Sandia National Laboratories nor
  the name of the author may be used to endorse or promote products
  derived from this software without specific prior written permission.

Warranty:
  This software is provided "as is" and without any express or
  implied warranties, including, without limitation, the implied
  warranties of merchantability and fitness for a particular
  purpose.

Package title:
  Out-of-core LU factorization for nCUBE (Vertex)

General Description:
  This code computes the LU factorization of a matrix on an nCUBE
  parallel computer running the Vertex operating system
  The matrices L is not available explicitly, but rather as a 
  product of Gauss matrices and permutation matrices.

Author: 
  David E. Womble
  Sandia National Laboratories, Dept. 1422
  P.O. Box 5800
  Albuquerque, NM 87185-5800

  dewombl@cs.sandia.gov
  (505) 845-7471

Author's conditions of release:
  1. The author received acknowledgement in any publication or in 
     any distributions of code in which this code is used
  2. The author is notified of any problems encountered with the
     code

Notes:
  Notes on the design, behavior and use of the program are in the
  README file.
*/

/* This is a development code for the matrix inversion package */

#include <math.h>
#include <stdio.h>
#include <fcntl.h>

#include "global_vars.h"
#include "macros.h"

main() {
  int buf[4];
  int mlen;
  int i, j, k, l;
  double *temp_seg;             /* temporary storage for a segment */
  int *bytes;                   /* bytes in each segment */
  int *low_bytes;               /* bytes in the lower tri portion of a seg */
  int my_rows_this, my_cols_this;
  int fd;                       /* file descriptor for disk IO */
  double *p;                    /* ptr to a segment for printing */
  double seconds();             /* function to generate timings */
  double timer0, timer1, timer2, timer3, timer4; 
  int rbytes, wbytes;           /* byte counters for I/O */

  whoami(&me, &me_proc, &host, &ndims_cube);
  nprocs_cube = 1<<ndims_cube;
  mask_cube = nprocs_cube-1;

  if (me == 0) {
    printf("Enter size of matrix: ");
    scanf("%d", &buf[0]);
    printf("Enter dimension of subcube to which each row is assigned: ");
    scanf("%d", &buf[1]);
    printf("Enter segment size: ");
    scanf("%d", &buf[2]);
    buf[3] = -1;
  }
  mlen = 4 * sizeof(int);
  fanout(me, mask_cube, buf, mlen);

  matrix_size = buf[0];
  nrows_matrix = matrix_size;
  ncols_matrix = matrix_size;

  ndims_row = buf[1];
  ndims_row = (ndims_row < ndims_cube) ? ndims_row : ndims_cube;
  ndims_col = ndims_cube - ndims_row;

  nprocs_col = 1<<ndims_col;
  nprocs_row = 1<<ndims_row;
  max_procs = (nprocs_row < nprocs_col) ? nprocs_col : nprocs_row;

  mask_row = nprocs_row - 1;
  mask_col = (nprocs_col - 1)<<ndims_row;

  my_first_col = inv_grey_c(me & mask_row);
  my_first_row = inv_grey_c((me & mask_col)>>ndims_row);

  my_rows = nrows_matrix/nprocs_col;
  if (my_first_row < nrows_matrix%nprocs_col) ++my_rows;
  my_cols = ncols_matrix/nprocs_row;
  if (my_first_col < ncols_matrix%nprocs_row) ++my_cols;

  ncols_seg = (buf[2] < ncols_matrix) ? buf[2] : ncols_matrix;
  ncols_seg = ((ncols_seg + max_procs - 1)/max_procs)*max_procs;

  nsegs_row = (ncols_matrix + ncols_seg - 1)/ncols_seg;
  my_cols_seg = ncols_seg/nprocs_row;
  ncols_last = ncols_matrix - ncols_seg*(nsegs_row - 1);
  my_cols_last = my_cols - my_cols_seg*(nsegs_row - 1);

  bytes_per_seg = (my_rows + 1)*my_cols_seg*sizeof(double);

  if (me == 0) {
    printf("\nMatrix size: %d x %d\n",nrows_matrix, ncols_matrix);
    printf("Processor deomposition is %d x %d\n",ndims_col, ndims_row);
    printf("Segment size is %d x %d\n", nrows_matrix, ncols_seg);
    printf("\n");
  }

  bytes = (int *) malloc(nsegs_row*sizeof(int));
  low_bytes = (int *) malloc(nsegs_row*sizeof(int));

  for (i=0; i<nsegs_row-1; i++) {
    bytes[i] = (my_rows + 1)*my_cols_seg*sizeof(double);
  }
  bytes[nsegs_row-1] = (my_rows + 1)*my_cols_last*sizeof(double);

  mat = (double **) malloc(4*sizeof(double *));
  for (i=0; i<4; i++){
    mat[i] = (double *) malloc(bytes[0]);
  }

  col1 = (double *) malloc((my_rows+1)*sizeof(double));
  col2 = (double *) malloc((my_rows+1)*sizeof(double));

  row1 = (double *) malloc((my_cols_seg+1)*sizeof(double));
  row2 = (double *) malloc((my_cols_seg+1)*sizeof(double));

  fd = create_file();

  timer0 = (double) 0.0;
  timer1 = (double) 0.0;
  wbytes = 0;

  timer0 = seconds(timer0);
  for (i = nsegs_row-1; i >= 0; i--) {
    init_seg(mat[0], i);
    timer1 = seconds(timer1);
    wbytes += put_seg(fd, mat[0], i, bytes, bytes[i]);
    timer1 = seconds(timer1);
    if (i == 1) {
      p = mat[0];
      mat[0] = mat[2];
      mat[2] = p;
    }
  }
  timer0 = seconds(timer0);

  if (me == 0) printf("Initialization time = "); timing(timer0, 2048, 0);
  if (me == 0) printf("Bytes written to disk in initialization = %d\n",wbytes);
  if (me == 0) printf("Initialization write time = "); timing(timer1, 2049, 0);
  
  if (me == 0) printf("\n");

  timer0 = (double) 0.0;
  timer1 = (double) 0.0;
  timer2 = (double) 0.0;
  timer3 = (double) 0.0;
  timer4 = (double) 0.0;

  wbytes = 0;
  rbytes = 0;
  timer0 = seconds(timer0);

  for (i = 0; i < nsegs_row; i++) {

    for (j = 0; j < i; j++) {
      /*
         Note: to overlap I/O with computation, uncomment the following
         section and change the calls to get_seg to calls to get_seg_end
         in the matching section of code (after the call to update-seg.

         The same thing can be done to the calls to get_seg immediatly
         before and after the call to factor_seg.
      */
      /*
      timer2 = seconds(timer2);
      if (i > 1) {
        if (j == i-1) {
          if (i < nsegs_row-1) {
            get_seg_st(fd, mat[2], 0, bytes, low_bytes[0]);
          }
        } else {
          get_seg_st(fd, mat[2], j+1, bytes, low_bytes[j+1]);
        }
      }
      timer2 = seconds(timer2);
      */

      if (me == 0) printf("Updating seg %d with seg %d\n",i,j);
      timer3 = seconds(timer3);
      update_seg(mat[1], j, mat[0], i);
      timer3 = seconds(timer3);

      timer2 = seconds(timer2);
      if (i > 1) {
        if (j == i-1) {
          if (i < nsegs_row-1) {
            rbytes += get_seg(fd, mat[2], 0, bytes, low_bytes[0]);
          }
        } else {
          rbytes += get_seg(fd, mat[2], j+1, bytes, low_bytes[j+1]);
        }
        p = mat[2];
        mat[2] = mat[1];
        mat[1] = p;
      }
      timer2 = seconds(timer2);
    }

    /*
    timer2 = seconds(timer2);
    if ((i != 0) && (i < nsegs_row-1)) {
      get_seg_st(fd, mat[2], i+1, bytes, bytes[i+1]);
    }
    timer2 = seconds(timer2);
    */

    if (me == 0) printf("Factoring seg %d\n",i);
    timer4 = seconds(timer4);
    factor_seg(mat[0], i);
    timer4 = seconds(timer4);

    timer2 = seconds(timer2);
    if ((i != 0) && (i < nsegs_row-1)) {
      rbytes += get_seg(fd, mat[2], i+1, bytes, bytes[i+1]);
    }
    timer2 = seconds(timer2);

    low_bytes[i] = perm_seg(mat[0], i, mat[3]);
    timer1 = seconds(timer1);
    wbytes += put_seg(fd, mat[3], i, bytes, bytes[i]);
    timer1 = seconds(timer1);

    if (i == 0) {
      p = mat[3];
      mat[3] = mat[1];
      mat[1] = p;
    }

    p = mat[2];
    mat[2] = mat[0];
    mat[0] = p;
  }
  timer0 = seconds(timer0);

  if (me == 0) printf("\n");
  if (me == 0) printf("Total run time = "); timing(timer0, 2050, 0);
  if (me == 0) printf("Bytes written to disk = %d\n",wbytes);
  if (me == 0) printf("Total write time = "); timing(timer1, 2051, 0);
  if (me == 0) printf("Bytes read from disk = %d\n",rbytes);
  if (me == 0) printf("Total read time = "); timing(timer2, 2052, 0);
  if (me == 0) printf("Total update time = "); timing(timer3, 2053, 0);
  if (me == 0) printf("Total factor time = "); timing(timer4, 2054, 0);

  /* Print statements for checking results */

  if (me == owner(nrows_matrix-1, ncols_matrix-1)) {
    p = mat[2] + (my_cols_last - 1)*(my_rows + 1) + my_rows - 1;
    printf("me = %d, Last entry = %f, last pivot = %f\n",me,*p,*(p+1));
  }

/*
  for (i=0; i<nsegs_row; i++){
    get_seg(fd, mat[0], i, bytes, bytes[i]);
    p = mat[0];
    my_cols_this = (i == nsegs_row-1) ? my_cols_last : my_cols_seg;
    for (k=0; k<(my_rows+1)*(my_cols_this); k++)
      printf("Node %d, (i,k) = (%d, %d), matrix entry = %f\n",
                me,i,k,*p++);
  }
*/

  close(fd);

  /* Deallocate memory. */

  for (i=0; i<4; i++){
    free(mat[i]);
  }
  free(row2);
  free(row1);
  free(col2);
  free(col1);
  free(mat);
  free(low_bytes);
  free(bytes);
}
