/*
Copyright (c) 1992 Sandia National Laboratories

  Redistribution and use in source and binary forms are permitted
  provided that source distributions retain this entire copyright
  notice.  Neither the name of Sandia National Laboratories nor
  the name of the author may be used to endorse or promote products
  derived from this software without specific prior written permission.

Warranty:
  This software is provided "as is" and without any express or
  implied warranties, including, without limitation, the implied
  warranties of merchantability and fitness for a particular
  purpose.

Package title:
  Out-of-core LU factorization for nCUBE (Vertex)

General Description:
  This code computes the LU factorization of a matrix on an nCUBE
  parallel computer running the Vertex operating system
  The matrices L is not available explicitly, but rather as a 
  product of Gauss matrices and permutation matrices.

Author: 
  David E. Womble
  Sandia National Laboratories, Dept. 1422
  P.O. Box 5800
  Albuquerque, NM 87185-5800

  dewombl@cs.sandia.gov
  (505) 845-7471

Author's conditions of release:
  1. The author received acknowledgement in any publication or in 
     any distributions of code in which this code is used
  2. The author is notified of any problems encountered with the
     code

Notes:
  Notes on the design, behavior and use of the program are in the
  README file.
*/

#include <math.h>
#include <stdio.h>
#include <fcntl.h>

perm_seg(fseg, fseg_num, tseg)
double *fseg;
int fseg_num;
double *tseg;
{
#include "extern_global_vars.h"
#include "macros.h"

  int j;                        /* loop counter */

  int ncols_this;               /* number of cols in this segment */
  int my_cols_this;             /* number of cols I own in this seg */

  int col_offset;               /* col number for first col in seg */
  int start_row;                /* row num to start column operations */
  int num_rows;                 /* number of rows in a column operation */

  double *ptr1,*ptr2;           /* running ptrs into segments */

  int counter;                  /* number of lower tri. entries */

  int one = 1;                  /* constant for the dcopy routines */

  if (fseg_num == (nsegs_row-1)) {
    ncols_this = ncols_last;
    my_cols_this = my_cols_last;
  } else {
    ncols_this = ncols_seg;
    my_cols_this = my_cols_seg;
  }
  counter = 0;

  col_offset = fseg_num*ncols_seg;
  ptr2 = tseg;

  for (j=0; j<ncols_this; j++) {
    if (me == col_owner(j)) {

      start_row = (col_offset + j)/nprocs_col;
      if (my_first_row <= j%nprocs_col) ++start_row;

      ptr1 = fseg + (j>>ndims_row)*(my_rows + 1) + start_row;

      num_rows = my_rows - start_row + 1;
      DCOPY(&num_rows, ptr1, &one, ptr2, &one);

      ptr2 += num_rows;
      counter += num_rows;
    }
  }

  for (j=0; j<ncols_this; j++) {
    if (me == col_owner(j)) {

      num_rows = (col_offset + j)/nprocs_col;
      if (my_first_row <= j%nprocs_col) ++num_rows;

      ptr1 = fseg + (j>>ndims_row)*(my_rows + 1);

      DCOPY(&num_rows, ptr1, &one, ptr2, &one);

      ptr2 += num_rows;
    }
  }

  return counter*sizeof(double);
}

copy_seg(fseg, fseg_num, tseg)
double *fseg;
int fseg_num;
double *tseg;
{
#include "extern_global_vars.h"
#include "macros.h"

  int my_cols_this;             /* number of cols I own in this seg */
  int num_ents;                 /* number of entries in the segment */
  int one = 1;                  /* constant for the dcopy routines */

  if (fseg_num == (nsegs_row-1)) {
    my_cols_this = my_cols_last;
  } else {
    my_cols_this = my_cols_seg;
  }

  num_ents = my_cols_this * (my_rows + 1);
  DCOPY(&num_ents, fseg, &one, tseg, &one);
}
  
