/*
Copyright (c) 1992 Sandia National Laboratories

  Redistribution and use in source and binary forms are permitted
  provided that source distributions retain this entire copyright
  notice.  Neither the name of Sandia National Laboratories nor
  the name of the author may be used to endorse or promote products
  derived from this software without specific prior written permission.

Warranty:
  This software is provided "as is" and without any express or
  implied warranties, including, without limitation, the implied
  warranties of merchantability and fitness for a particular
  purpose.

Package title:
  Out-of-core LU factorization for nCUBE (Vertex)

General Description:
  This code computes the LU factorization of a matrix on an nCUBE
  parallel computer running the Vertex operating system
  The matrices L is not available explicitly, but rather as a 
  product of Gauss matrices and permutation matrices.

Author: 
  David E. Womble
  Sandia National Laboratories, Dept. 1422
  P.O. Box 5800
  Albuquerque, NM 87185-5800

  dewombl@cs.sandia.gov
  (505) 845-7471

Author's conditions of release:
  1. The author received acknowledgement in any publication or in 
     any distributions of code in which this code is used
  2. The author is notified of any problems encountered with the
     code

Notes:
  Notes on the design, behavior and use of the program are in the
  README file.
*/

#include <math.h>
#include <stdio.h>
#include <fcntl.h>

update_seg(fseg, fseg_num, tseg, tseg_num)
double *fseg;
int fseg_num;
double *tseg;
int tseg_num;
{
#include "extern_global_vars.h"
#include "macros.h"

  int i, j, l, k;               /* loop counters */

  int ncols_this;               /* number of cols in this segment */
  int my_cols_this;             /* number of cols I own in this seg */

  struct pivot_type {           /* a stucture for storing pivot info */
    double entry;               /*   pivot entry */
    int row;                    /*   pivit row number */
  } pivot, tpivot;              /* pivot info and some temporary pivot infor */
  double pivot_row_elem;        /* element of pivot row to use in elim. */

  int bc_start;                 /* starting row for broadcasts */
  int up_start;                 /* starting row for updates */
  int num_rows;                 /* number of rows in a column operation */
  int num_cols;                 /* number of cols in a row operation */

  int pivot_row_in_proc;        /* local pivot row for proc containing pivot */
  int col_offset;               /* column number for 0-th column in from seg */

  double *ptr1,*ptr2;           /* running ptrs into mat and update cols */
  double *ptr3, *ptr4;          /* running ptrs to diagonal and pivot rows */
  double *ptr5;                 /* running ptr into fseg */

  int col2_avail;               /* flag to indicate that col2 is avail. */

  int bytes;                    /* number of bytes in messages */
  int root;                     /* root processor for fanout */
  int type;                     /* mesage type for messages */
  int dest;                     /* destination processor for a single mess. */
  int flag;                     /* flag for messages */

  int one = 1;                  /* constant for the daxpy routines */
  int zero = 0;                 /* constant for the daxpy routines */

  int stride;                   /* stride for vector copy routines */
  double scaler;                /* scaler for daxpy routines */

  int error_num;                /* errors from ncube system routines */

  if (tseg_num == (nsegs_row-1)) {
    ncols_this = ncols_last;
    my_cols_this = my_cols_last;
  }
  else {
    ncols_this = ncols_seg;
    my_cols_this = my_cols_seg;
  }

  col2_avail = -1;

  ptr5 = fseg;

  for (j=0; j<=ncols_seg; j++) {

    col_offset = fseg_num*ncols_seg;
    up_start = bc_start; /* lag by one iteration */
    bc_start = (col_offset + j)/nprocs_col;
    if (my_first_row <= j%nprocs_col) ++bc_start;

    ptr4 = row2;

    if (j != ncols_seg){
      if (me == col_owner(j)) {

        /* now broadcast update column to other processors */

        num_rows = my_rows - bc_start + 1;
        bytes = num_rows*sizeof(double);
        root = col_owner(j);
        type = 100 + fseg_num;
        fanout2(me, mask_row, ptr5, bytes, root, type);
        ptr2 = col2 + bc_start;
        DCOPY(&num_rows, ptr5, &one, ptr2, &one);

        ptr5 += num_rows;
        col2_avail = 1;
      }
    }

    if (j != 0) {
      for (l=0; l<my_cols_this; l++) {
        /* update col l with col j-1 */
        ptr1 = tseg + l*(my_rows + 1) + up_start;
        ptr2 = col1 + up_start;
        num_rows = my_rows - up_start;
        scaler = -*ptr4;
        DAXPY(&num_rows, &scaler, ptr2, &one, ptr1, &one);
        ++ptr4;
      }
    }

    /* If this processor is not responsible for column j, then it
       has to receive the column to use as an update */

    if (j != ncols_seg){
      if (me != col_owner(j)){
        bytes = (my_rows - bc_start + 1)*sizeof(double);
        root = col_owner(j);
        type = 100 + fseg_num;
        ptr2 = col2 + bc_start;
        fanout2(me, mask_row, ptr2, bytes, root, type);

        col2_avail = 1;
      }

      /* make sure information has been received and set the right
         pointer to it */

      if (col2_avail > 0){
        ptr2 = col1;
        col1 = col2;
        col2 = ptr2;

        col2_avail = -1;
      } else {
        printf("Node %d error -- update column not available\n",me);
      }

      /* exchange rows using information in last row of pivot column */

      pivot.row = *(col1 + my_rows);
      pivot_row_in_proc = pivot.row/nprocs_col;
      num_cols = my_cols_this;
      if (pivot.row != (col_offset + j)){
        if (my_first_row == (j%nprocs_col)){
          ptr3 = tseg + (col_offset+j)/nprocs_col;
          stride = my_rows + 1;
          DCOPY(&num_cols, ptr3, &stride, row1, &one);
          if (my_first_row != pivot.row%nprocs_col) {
            dest = row_owner(pivot.row);
            bytes = num_cols*sizeof(double);
            type = 500;
            nwrite(row1, bytes, dest, type, &flag);
          }
        }
      }
      if (my_first_row == pivot.row%nprocs_col){
        ptr3 = tseg + pivot_row_in_proc;
        stride = my_rows + 1;
        DCOPY(&num_cols, ptr3, &stride, row2, &one);
      }
      /* fanout pivot row */
      root = row_owner(pivot.row);
      bytes = num_cols*sizeof(double);
      type = 400;
      fanout2(me, mask_col, row2, bytes, root, type);
      if (pivot.row != (col_offset + j)){
        if (my_first_row == (pivot.row%nprocs_col)){
          if (my_first_row != (j%nprocs_col)) {
            dest = row_owner(j);
            bytes = num_cols*sizeof(double);
            type = 500;
            nread(row1, bytes, &dest, &type, &flag);
          }
          ptr3 = tseg + pivot_row_in_proc;
          stride = my_rows + 1;
          DCOPY(&num_cols, row1, &one, ptr3, &stride);
        }
        if (my_first_row == (j%nprocs_col)) {
          ptr3 = tseg + (col_offset+j)/nprocs_col;
          stride = my_rows + 1;
          DCOPY(&num_cols, row2, &one, ptr3, &stride);
        }
      }
    }
  }
}
